TREE_IDENTIFIER = "tree_id"


class GRAPH:
    class ATTRIBUTES:
        CREATED_AT = "created_at"

    class LABELS:
        MENTIONS = "mentions"
        PARENT_OF = "parent_of"
        AUTHOR_OF = "author_of"
        ANSWERED_BY = "answered_by"

    class SUBSETS:
        TWEETS = "tweets"
        AUTHORS = "authors"


class TABLE:
    class COLUMNS:
        PARENT_ID = "parent_id"
        CREATED_AT = "created_at"
        AUTHOR_ID = "author_id"
        TEXT = "text"
        POST_ID = "post_id"
        TREE_ID = "tree_id"
